/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#ifndef MATHLIB_H
#define MATHLIB_H

#include <math.h>

#ifndef PI
 #define PI		3.14159265358979323846264338328
#endif
#ifndef PI2
 #define PI2		6.28318530717958647692528676656		/* 2*PI */
#endif
#define PIOVER2		1.57079632679489661923132169164		/* PI/2 */
#define PI3OVER2	4.71238898038468985769396507492		/* 3*PI/2 */
#define ONEOVERPI	0.31830988618379067153776752674		/* 1/PI */
#define TWOOVERPI	0.63661977236758134307553505349		/* 2/PI */

/* it seems that rint is not completely standard. */
#define	round( d )	((int)rint( d ))

struct u_vector {
	fFloat			x, y, z;
};
typedef struct u_vector u_Vector;

struct u_quaternion {
	fFloat			s;
	u_Vector		v;
};
typedef struct u_quaternion u_Quaternion;

/* a 3x3 matrix to change basis */
typedef fFloat u_Matrix33[3][3];

typedef fFloat u_Matrix44[4][4];

struct u_plane {
	u_Vector 		normal;
	fFloat 			dist; 	/* shortest distance from origin, world coord */
	fFloat 			a, b, c, d;
	u_Vector		p;	/* a point in the plane */	
};
typedef struct u_plane u_Plane;

struct u_Sphere {
	fFloat			radius;
	u_Vector		center;
};
typedef struct u_sphere u_Sphere;

/* vector.c */
extern void 		initV 
		( u_Vector *vect, fFloat x, fFloat y, fFloat z );
extern void 		zeroV 
		( u_Vector *vect );
extern fFloat		normV
		( u_Vector *vect );
extern fFloat		sqnormV 
		( u_Vector *vect );
extern void 		copyV
		( u_Vector *target, const u_Vector *src );
extern void 		addV
		( u_Vector *vect1, u_Vector *vect2, u_Vector *result );
extern void 		subV
		( u_Vector *vect1, u_Vector *vect2, u_Vector *result );
extern void 		scalarMV
		( u_Vector *vect, fFloat scalar, u_Vector *result );
extern void 		normalizeV
		( u_Vector *vect );
extern fFloat		 dotprodV
		( u_Vector *vect1, u_Vector *vect2 );
extern void		 crossprodV
		( u_Vector *vect1, u_Vector *vect2, u_Vector *result );
extern fFloat 		dist
		( u_Vector *v1, u_Vector *v2 );
extern fFloat		 sqdist
		( u_Vector *v1, u_Vector *v2 );
extern void		 multM3
		( u_Matrix33 a, u_Matrix33 b, u_Matrix33 c );
extern void 		invertbasis
		( u_Matrix33 orig, u_Matrix33 inv );
extern void 		changebasis
		( u_Vector *v, u_Matrix33 b, u_Vector *r );
extern void		 basisVA
		( u_Vector *v, fFloat angle, u_Matrix33 m );

/* sphere.c */
extern bool		u_sphereOverlap
		( const u_Sphere *s, const u_Sphere *t );
extern bool		u_vectorinSphere
		( const Sphere *s, const u_Vector *v );

/* quat.c */
extern void		initQuatD 
		( u_Quaternion *q, fFloat s, fFloat x, fFloat y, fFloat z );
extern void		initQuatVS 
		( u_Quaternion *q, fFloat s, u_Vector *v );
extern void 		initQuatVA 
		( u_Quaternion *q, fFloat angle, u_Vector *v );
extern void 		zeroQ 
		( u_Quaternion *q );
extern void 		copyQ 
		( u_Quaternion *dest, u_Quaternion *src );
extern dfloat 		normQ
		( u_Quaternion *q );
extern void		normalizeQ 
		( u_Quaternion *q, u_Quaternion *r );
extern void		conjugateQ
		( u_Quaternion *q, u_Quaternion *r );
extern void 		addQ 
		( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r );
extern void 		subQ 
		( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r );
extern void 		scalarMQ
		( u_Quaternion *q, fFloat m, u_Quaternion *r);
extern void		multQ 
		( u_Quaternion *q1, u_Quaternion *q2, u_Quaternion *r );
extern void		inverseQ 
		( u_Quaternion *q1, u_Quaternion *r );
extern void		qvq 
		( u_Quaternion *q1, u_Vector *v, u_Quaternion *q2, u_Vector *r );

/* linsys.c */
#define linsys( a, b, c, d )	gausselim( a, b, c, d )
extern sInt 		gausselim
		( sInt n, dfloat A[n][n], dfloat B[n], dfloat x[n] );

/* matrix.c */
extern void		identity
		( MATRIX_UL a );
extern void 		addM 
		( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r );
extern void 		subM 
		( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r );
extern void 		scalarmM 
		( MATRIX_UL a, const double m, MATRIX_UL r );
extern void 		multM 
		( MATRIX_UL a, MATRIX_UL b, MATRIX_UL r );
extern void 		multVM 
		( MATRIX_UL m, u_Vector *v, u_Vector *result );

/* analytic.c */
extern void initPlaneP ( PLANE_UL *p, POINT_UL *p1, POINT_UL *p2, POINT_UL *p3 );
extern void initPlaneVD ( PLANE_UL *p, u_Vector *v, double d );
extern int interPlaneLine ( PLANE_UL *p, u_Vector *v1, u_Vector *v2,
		 u_Vector *r );

/* trig.c */
extern double TrigTable[];
extern void initTrigTable ( void );
extern double sinus ( double rad );
extern double sinusd ( double deg );
extern double cosinus ( double rad );
extern double cosinusd (double deg );

#endif
 
