/* Unreality
 * Copyright (C) 2000 Bruno Barberi Gnecco
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, go to <http://www.gnu.org>.
 */

#include "mathlib.h"

/* 
 * 3x3 Matrices
 */

/* assumes b != c != a */
void multM3 ( u_Matrix33 a, u_Matrix33 b, u_Matrix33 c ) {
	sInt			i, j, k;

	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ ) {
			c[i][j] = 0;
			for ( k = 0; k < 3; k++ )
				c[i][j] += a[i][k]*b[k][j];
		}

}

#ifdef TO_BE_REVIEWED
/* if bases are orthonormal, then the basis change matrix is orthogonal,
because it's only a rotation. From this follows that the inverse is the
transpose. */
void invertbasis( u_Matrix33 orig, u_Matrix33 inv ) {
	sInt			i, j;
	u_Matrix33 		b;

	if ( &orig[0][0] != &inv[0][0] )
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				inv[j][i] = orig[i][j];
	else {
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				b[j][i] = orig[i][j];
		for ( i = 0; i < 3; i++ )
			for ( j = 0; j < 3; j++ )
				inv[i][j] = b[i][j];
	}
} 

void changebasis ( u_Vector *v, u_Matrix33 b, u_Vector *r ) {
	if ( v != r ) {
		r->x = b[0][0] * v->x + b[0][1] * v->y + b[0][2] * v->z;
		r->y = b[1][0] * v->x + b[1][1] * v->y + b[1][2] * v->z;
		r->z = b[2][0] * v->x + b[2][1] * v->y + b[2][2] * v->z;
		return;
	}
	else {
		prv.x = b[0][0] * v->x + b[0][1] * v->y + b[0][2] * v->z;
		prv.y = b[1][0] * v->x + b[1][1] * v->y + b[1][2] * v->z;
		prv.z = b[2][0] * v->x + b[2][1] * v->y + b[2][2] * v->z;
		copyV( r, &prv );
	}
}

/* given a vector and an angle, this returns the basis change matrix from
canonical to it */
/***********SHOULD BE TESTED */
void basisVA ( u_Vector *v, fFloat angle, u_Matrix33 m ) {
	fFloat			a, b, d;
	u_Vector		fy;
	u_Quaternion		rot, invrot;
	
	/* new basis (let's call f) x component is v, normalized */
	copyV( &prv, v );
	normalizeV( &prv );
	m[0][0] = prv.x;	m[0][1] = prv.y;	m[0][2] = prv.z;
	d = sqrt(prv.x*prv.x + prv.y*prv.y );
	a = prv.y / d;
	b = prv.x / d;
	if ( prv.x > 0 && prv.y > 0 ) {
		a *= -1;
	}
	if ( prv.x > 0 && prv.y < 0 ) {
		a *= -1;
		b *= -1;
	} 
	if ( prv.x < 0 && prv.y > 0 ) {
		b *= -1;
	}
	/* this is fy */
	fy.x = a;	      /* these values */
	fy.y = b;	      /* are for */
	fy.z = 0;	      /* angle = 0 */
	
	/* now rotate */
	if ( angle ) {
		initQuatVA( &rot, angle, &prv );
		conjugateQ( &rot, &invrot );
		qvq( &rot, &fy, &invrot, &fy );
	}
	normalizeV( &fy );
	m[1][0] = fy.x;		m[1][1] = fy.y;		m[1][2] = fy.z;
	
	crossprodV( &prv, &fy, &prv );
	normalizeV( &prv );	/* it should be normalized, but it doesn't hurt */
	m[2][0] = prv.x;	m[2][1] = prv.y;	m[2][2] = prv.z;
}
#endif

/*
 * 4x4 Matrices
 */

static u_Matrix44		prm;
static u_Vector			prv;

static const u_Matrix44 	ident = { { 1.0, 0.0, 0.0, 0.0 },
					  { 0.0, 1.0, 0.0, 0.0 },
					  { 0.0, 0.0, 1.0, 0.0 },
					  { 0.0, 0.0, 0.0, 1.0 } };

void identity ( u_Matrix44 a ) {
	memcpy( a, ident, sizeof(fFloat)<<4 );
}

void copyM ( u_Matrix44 target, u_Matrix44 src ) {
	memcpy( target, src, sizeof(fFloat)<<4 );
}

void addm ( u_Matrix44 a, u_Matrix44 b, u_Matrix44 r ) {
	sInt			i, j;

	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j] + b[i][j];
}

void subm ( u_Matrix44 a, u_Matrix44 b, u_Matrix44 r ) {
	sInt			i, j;

	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j] - b[i][j];
}

void scalarmM ( u_Matrix44 a, const double m, u_Matrix44 r ) {
	sInt			i, j;

	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++ )
			r[i][j] = a[i][j]*m;
}

void multM ( u_Matrix44 a, u_Matrix44 b, u_Matrix44 r ) {
	sInt			i, j, t = 0;
	fFloat *p;
	
	if ( &a[0][0] != &r[0][0] && &b[0][0] != &r[0][0] ) {
		p = r[0];
		t = 1;
	}
	else
		p = prm[0];
	
	for ( i = 0; i < 4; i++ )
		for ( j = 0; j < 4; j++, p++ )
			*p = a[i][0]*b[0][j] +
			     a[i][1]*b[1][j] +
			     a[i][2]*b[2][j] +
			     a[i][3]*b[3][j];
	if ( t == 0 )
		copyM( r, prm );
}

void multVM ( u_Matrix44 m, u_Vector *v, u_Vector *result ) {
	u_Vector			*p;

	if ( v != result ) {
		result->x = v->x*m[0][0] + v->y*m[1][0] + v->z*m[2][0] + m[3][0];
		result->y = v->x*m[0][1] + v->y*m[1][1] + v->z*m[2][1] + m[3][1];
		result->z = v->x*m[0][2] + v->y*m[1][2] + v->z*m[2][2] + m[3][2];
	}
	else {
		prv.x = v->x*m[0][0] + v->y*m[1][0] + v->z*m[2][0] + m[3][0];
		prv.y = v->x*m[0][1] + v->y*m[1][1] + v->z*m[2][1] + m[3][1];
		prv.z = v->x*m[0][2] + v->y*m[1][2] + v->z*m[2][2] + m[3][2];
		copyV( result, &prv );
	}
}
